/* 
 * File:   main.cpp
 * Author: Will Matheson
 * Course: PROG 2400 (Data Structures), Section 700
 * Due:    April 18, 2013
 * 
 * PROG 2400 Assignment #4: Sorting
 * Part 2: External Merge Sort
 * 
 * Performs a stable external merge sort on 10,000 numbers loaded from a file.
 * Ten merge sorts of 1,000 numbers each are performed and the results are
 * written to two files. Finally, the two files are read in ten concurrent
 * places and a new file of 10,000 sorted numbers is built.
 *
 * Created on April 10, 2013, 5:56 PM
 */

#include <cstdlib>
#include <iostream>
#include <fstream>
#include <string>
#include <cstring>
#include <sstream>
#include <vector>
#include <limits>
#include "Timer.h"

using namespace std;

//work along a line to split it into delimited pieces
vector<string> &split(const string &s, char delim, vector<string> &elems) {
    stringstream ss(s);
    string item;
    while (getline(ss, item, delim)) {
        elems.push_back(item);
    }
    return elems;
}

//helper function to split a string by a delimiter
//http://stackoverflow.com/a/236803
vector<string> split(const string &s, char delim) {
    vector<string> elems;
    split(s, delim, elems);
    return elems;
}

//return smaller of two numbers - if equal, return second number
int min(int x, int y)
{
    if (x < y)
    {
        return x;    
    }
    else
    {
        return y;    
    }
}

void mergeSortRecurs(int numbers[], string lines[], int left, int right,
        int* scratchInt, string* scratchString){
    //a one-element sublist is definitely sorted - we're done here!
    if (right == left+1)
    {
        return;            
    }
    else
    {
        int length = right - left;
        int midpointDistance = length/2;
        //l and r are the positions in the L and R subarrays
        int l = left, r = left + midpointDistance;
        
        //sort subarrays
        mergeSortRecurs(numbers, lines, left, left + midpointDistance,
                scratchInt, scratchString);
        mergeSortRecurs(numbers, lines, left + midpointDistance, right,
                scratchInt, scratchString);
        
        //merge arrays together and use "scratch" for temporary storage
        for(int i=0;i<length;i++)
        {
            //any elements in the left array?
            //if so, check if any elements in right array...
            //if so, we compare them and see if left is less or equal
            //otherwise just take the next element from the right array
            if ((l < left + midpointDistance) &&
                    ((r == right)
                        || (min(numbers[l], numbers[r]) == numbers[l])))
            {
                scratchInt[i] = numbers[l];
                scratchString[i] = lines[l];
                l++;            
            }
            else
            {
                scratchInt[i] = numbers[r];
                scratchString[i] = lines[r];
                r++;            
            }
        }
        
        //copy sorted array from "scratch" back to the base array
        for (int i = left; i<right; i++)
        {
            numbers[i] = scratchInt[i - left];
            lines[i] = scratchString[i - left];
        }
    }
}

//controls the recursive merge sort
//http://www.cprogramming.com/tutorial/computersciencetheory/merge.html
bool mergeSortWrapper(int numbers[], string lines[], int blockSize){
    //allocate "scratch space"
    int* scratchInt = (int*)malloc(blockSize * sizeof(int));
    string* scratchString = new string[blockSize];
    if ((scratchInt != NULL) && (scratchString != NULL))
    {
        //start the recursive part
        mergeSortRecurs(numbers, lines, 0, blockSize,
                scratchInt, scratchString);
        
        //free the memory
        free(scratchInt);
        return true;
    }
    else
    {
        cerr << "No memory available for sorting, or invalid block size."
                << endl;
        return false;        
    }
}

//coordinates the 1000-at-a-time merge sorts
bool chunksSort(string fileNames[], int numbers[], string lines[],
        int blockSize, int totalSize){
    //set up input file
    ifstream inStream(fileNames[0].c_str());//takes old-school string
    if (inStream.fail())
    {
        cerr << "Couldn't open numbers file." << endl
                << "Does the file exist? Do you have access privileges here?" 
                << endl;
        return false;                
    }
    
    //set up output files
    ofstream outStreamA(fileNames[1].c_str());
    ofstream outStreamB(fileNames[2].c_str());
    if(outStreamA.fail() || outStreamB.fail())
    {
	cerr << "Couldn't open output stream." << endl
                << "Do you have write privileges here? Is the disk full?" 
                << endl;
        return false;
    }
    
    //load up 10 sets of 1000 numbers
    for (int i=0; i<totalSize; i+=blockSize)
    {
        //load up 1000 numbers
        for (int j=0; j<blockSize; j++)
        {
            //load lines from original file
            string fileLine;
            getline(inStream,fileLine);

            //split on comma into two parts
            vector<string> lineParts = split(fileLine,',');

            //line number
            string lineNumber = lineParts.at(0);
            lines[j] = lineNumber;

            //number itself
            stringstream ss(lineParts.at(1));
            int number;
            ss >> number;
            numbers[j] = number;
        }

        //do the sort
        if (!mergeSortWrapper(numbers, lines, blockSize))
        {
            cerr << "Sorting aborted." << endl;
            return false;            
        }

        //put the sorted array into the scratch file
        //first half of them go to the first file
        if (i<(totalSize/2))
        {
            for (int j=0; j<blockSize; j++)
            {
                outStreamA << lines[j] << "," << numbers[j];               
                if ((i + j) < (totalSize / 2))
                {
                    outStreamA << endl;                    
                }
            }
        }
        //second half go to the second file
        else
        {
            for (int j=0; j<blockSize; j++)
            {
                outStreamB << lines[j] << "," << numbers[j];               
                if ((i + j) < (totalSize))
                {
                    outStreamB << endl;                    
                }
            }            
        }
    }
    
    //close the streams
    inStream.close();  
    outStreamA.close();
    outStreamB.close();
    return true;

}

//returns false if not all of the counters have advanced past the max yet
bool counterCompletionCheck(int counters[], int numCounters, int maxNum){
    for (int i=0; i<numCounters; i++)
    {
        if (counters[i] <= maxNum)
        {
            return false;                    
        }
    }
    return true;
}

//merge n sorted lists from two files into new file
bool nWayMerge(string fileNames[], int nWays, int blockSize){
    int counters[nWays];//counters for each cursor (doing 1000 reads each)
    string lines[nWays];//the entire current line of each cursor
    int numbers[nWays];//the "current number" for each "cursor"

    //this is just here so that we can track occurrences of triples (or better)
    vector<int> lastThree;
    //start with 3 nonsense values to simplify code that checks last 3 values
    lastThree.push_back(-1);
    lastThree.push_back(-2);
    lastThree.push_back(-3);

    fstream inStreams[nWays];//to read the two files at (for us) 10 places
    ofstream outStream(fileNames[3].c_str());//to write a new file as we go
    if(outStream.fail())
    {
	cerr << "Couldn't open output stream." << endl
                << "Do you have write privileges here? Is the disk full?" 
                << endl;
        return false;
    }
    //set up the first half of the streams to read from the first file
    for (int i=0; i<(nWays/2); i++)
    {
        inStreams[i].open(fileNames[1].c_str(), ios::in);
        if (inStreams[i].fail())
        {
            cerr << "Couldn't open input stream number " << i + 1 << "." << endl
                   << "Does the file exist? Do you have access privileges here?" 
                   << endl;
            return false;
        }
        //go to appropriate line (http://stackoverflow.com/a/5207600/1736461)
        for(int j=0; j < (i*blockSize); ++j)
        {
            inStreams[i].ignore(numeric_limits<streamsize>::max(),'\n');
        }
        counters[i] = 0;
    }
    //set up second half to read from second file
    for (int i=(nWays/2); i<nWays; i++)
    {
        inStreams[i].open(fileNames[2].c_str(), ios::in);
        for(int j=0; j < ((i-(nWays/2))*blockSize); ++j)
        {
            inStreams[i].ignore(numeric_limits<streamsize>::max(),'\n');
        }
        counters[i] = 0;
    }
    
    //get the first piece of data from each stream
    for (int i=0; i<nWays; i++)
    {
        getline(inStreams[i], lines[i]);
        //get number from file
        vector<string> lineParts = split(lines[i],',');
        stringstream ss(lineParts.at(1));
        ss >> numbers[i];
        counters[i]++;
    }

    //http://en.wikipedia.org/wiki/Merge_algorithm
    while (!counterCompletionCheck(counters, nWays, blockSize))
        //while anybody still points to data
    {
        //find the leftmost lowest number
        int lowKeyCursor = 0;//always start from left col (lowest line numbers)
        for (int i=0; i<nWays; i++)
        {
            //only pick a cursor to the right if it's lower
            if (numbers[i] < numbers[lowKeyCursor])
            {
                lowKeyCursor = i;
            }
        }

        //chuck it on the pile
        outStream << lines[lowKeyCursor] << endl;
        //advance our counter for that cursor
        counters[lowKeyCursor]++;
        
        //check for triples (output the value so we can find it later)
        //if there were a run of four or more, you'd see the same message twice
        lastThree.insert(lastThree.begin(), numbers[lowKeyCursor]);
        if ((lastThree.at(0) == lastThree.at(1))
                && (lastThree.at(0) == lastThree.at(2)))
        {
            cout << "TRIPLE! " << numbers[lowKeyCursor] << endl;        
        }
        
        //get next line for that cursor, if it's eligible
        if (counters [lowKeyCursor] <= blockSize)
        {
            getline(inStreams[lowKeyCursor], lines[lowKeyCursor]);
            vector<string> lineParts = split(lines[lowKeyCursor],',');
            stringstream ss(lineParts.at(1));
            ss >> numbers[lowKeyCursor];
        }
        else
        {
            //fling it out of range so it won't ever be 'low' anymore
            numbers[lowKeyCursor] = numeric_limits<int>::max();
        }
    }    
    return true;
}

//entry point of program
int main(int argc, char** argv) {
    //we're going to load and sort 10,000 random numbers btw. 0 and 100,000
    const int totalSize = 10000;
    //...in chunks of 1000
    const int blockSize = 1000;
    //so at the end there'll be a 10-way merge
    int nWays = totalSize / blockSize;
    
    //array to hold the blocks of numbers
    static int numbers[blockSize];
    //array to hold the original line numbers (along for the ride)
    static string lines[blockSize];
    //file to read from, first scratch file, second scratch file, output file
    static string fileNames[]
         = {"usortnum.txt", "filea", "fileb", "sortnum.txt"};
    
    cout << "Reading numbers from " << fileNames[0] << "." << endl;
    //start timer
    Timer* myTimer = new Timer();
    myTimer->start();
    
    //sort from original file in chunks of 1000, write to scratch files
    if (!chunksSort(fileNames, numbers, lines, blockSize, totalSize))
    {
        cerr << "Aborting. Sorting sublists incomplete." << endl;
        return -1;
    }    
    
    //merge resulting sorted lists from scratch files into new file
    if (!nWayMerge(fileNames, nWays, blockSize))
    {
        cerr << "Aborting. Writing new sorted file incomplete." << endl;
        return -1;    
    }
    
    //report time
    unsigned long elapsedTime = myTimer->elapsedTimeMilliseconds();
    delete myTimer;
    cout << "External merge sort took " << elapsedTime << " ms." << endl;
    cout << "Results written to " << fileNames[3] << "." << endl;
    
    return 0;
}
