/* 
 * File:   LinkedList.cpp
 * Author: Will Matheson
 * Course: PROG 2400 (Data Structures), Section 700
 * Due:    March 25, 2013
 * 
 * Implementation for LinkedList (see header file for description)
 * 
 * Based on earlier LinkedList, which was created January 11, 2013, 3:32 PM
 */

#include "LinkedList.h"

//default constructor
LinkedList::LinkedList() {
    head = NULL;//a train without any cars
}

//copy constructor (not defined)
LinkedList::LinkedList(const LinkedList& orig) {
}

//virtual destructor (not defined)
LinkedList::~LinkedList() {
}

//add a line at the end of the LinkedList
void LinkedList::Append(int in){
    ListNode* newNode = new ListNode();//create new node
    newNode->setLine(in);//set its data
    if (head == NULL)//we're just starting up the chain
    {
        head = newNode;
    }
    else //we're adding onto the chain, so find the end of the chain
    {
        ListNode* runner = head;
        ListNode* trailer = NULL;
        while(runner != NULL) //while he hasn't fallen off the back of the train
        {
            //move onto the next node
            trailer = runner;
            runner = runner->getNext();
        }
        
        //now, attach the new car to the end of the train
        trailer->setNext(newNode);
    }
}

//pivot link
LinkedList* LinkedList::Concat(LinkedList* firstList, int pivotValue, LinkedList* secondList){
    LinkedList* newList = new LinkedList();
    for (int i=0; i<firstList->Count(); i++)
    {
        newList->Append(firstList->ValueAt(i));    
    }
    newList->Append(pivotValue);
    for (int i=0; i<secondList->Count(); i++)
    {
        newList->Append(secondList->ValueAt(i));    
    }
    delete firstList;
    delete secondList;
}

//value at
int LinkedList::ValueAt(int index){
  
    ListNode* runner = head;
    
    if (index < 0)
    {
        std::cerr << "Line numbers must be non-negative integers." << std::endl;
        return -1;
    }
    else
    {
        //go to the nth spot (middle or end of list)
        for (int i=0; i<index; i++)
        {
            runner = runner->getNext();
            if (runner == NULL)
            {
                std::cerr << "Line specified for listing does not exist."
                        << std::endl;
                return -1;
            }
        }
        
        //display the first line
        return runner->getLine();
        
    }
}    

//delete a line by its index (the LinkedList is 0-indexed)
void LinkedList::Delete(int index){
    ListNode* runner = head;
    ListNode* trailer = NULL;
    if (index < 0)
    {
        std::cerr << "Line must be a non-negative integer." << std::endl;
        return;
    }
    //if index is 0 (start of list)
    else if (index == 0)
        head = runner->getNext();
    else
    {
        //go to the nth spot (middle or end of list)
        for (int i=0; i<index; i++)
        {
            trailer = runner;
            runner = runner->getNext();
            if (runner == NULL)
            {
                std::cerr << "Line specified for deletion does not exist."
                        << std::endl;
                return;
            }
        }
        
        //"skip" car to be deleted
        trailer->setNext(runner->getNext());
    }    
    
    delete runner;//delete the node from memory
}

//insert a line at an index (the LinkedList is 0-indexed)
void LinkedList::Insert(int in, int index){
    index ++;//treat the first line as line 1 so we can still insert when empty
    ListNode* newNode = new ListNode();//new node to hold the line
    newNode->setLine(in);//put the line into that new node
    
    ListNode* runner = head;
    ListNode* trailer = NULL;
    int counter = 0;//for keeping track of where we are
    
    while(runner != NULL)
    {
        //increase counter
        //check if it is location sought
        if (++counter == index)
            break;//stop moving ahead
        trailer = runner;
        runner = runner->getNext();
    }
    
    if(runner != NULL)
    {
        if(trailer == NULL)
        {
            //inserting at beginning
            newNode->setNext(head);
            head = newNode;
        }
        else
        {
            //inserting in middle
            //we can't "insert" something to be on the end, use append for that
            newNode->setNext(trailer->getNext());
            trailer->setNext(newNode);                
        }    
    }
    else
    {
        //location doesn't exist
        std::cerr << "Insertion location out of bounds of document." 
                << std::endl;
        return;
    }

}

//count the stored lines
int LinkedList::Count(){
    int lines = 0;
    ListNode* node = head;
    while (node != NULL)//while we haven't fallen off the back of the train
    {
        node = node->getNext();
        lines ++;
    }
    return lines;
}

